class ConfigException(Exception):
    pass


class InvalidDatetimeInConfig(ConfigException):
    pass


class CouldNotAlignException(Exception):
    pass


class InternightException(Exception):
    pass
